# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END

import copy
import logging
import json
import os
from foglamp.common import logger
import filter_ingest
from foglamp.plugins.filter.amber.client import AmberClient

__author__ = "Ashish Jabble, Mark Riddoch"
__copyright__ = "Copyright (c) 2024 Dianomic Systems Inc."
__license__ = "Apache 2.0"

_LOGGER = logger.setup(__name__, level=logging.WARN)

PLUGIN_NAME = "amber"

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Filter to send data via the Amber API to the Boon Logic Clustering Engine.',
        'type': 'string',
        'default': PLUGIN_NAME,
        'readonly': 'true'
    },
    'modelName': {
        'description': 'A name to give the model. This is merely used as a prefix to the sensors created in the filter.',
        'type': 'string',
        'default': 'amber',
        'order': '2',
        'group': 'Naming',
        'displayName': 'Model Name'
    },
    'learningSamples': {
        'description': 'Number of samples required to switch from learning to monitoring mode. This includes the number of samples required for auto tuning.',
        'type': 'integer',
        'default': '100000',
        'minimum': '1000',
        'order': '3',
        'group': 'Model Tuning',
        'displayName': 'Learning Samples'
    },
    'autoTuningSamples': {
        'description': 'Number of auto tuning samples.',
        'type': 'integer',
        'default': '10000',
        'minimum': '100',
        'order': '4',
        'group': 'Model Tuning',
        'displayName': 'Auto Tuning Samples'
    },
    'learningRateNumerator': {
        'description': 'A model transitions from learning to monitoring mode if fewer than this number of  new clusters are opened in the last Learning Rate Denominator samples.',
        'type': 'integer',
        'default': '10',
        'minimum': '1',
        'order': '5',
        'group': 'Model Tuning',
        'displayName': 'Learning Rate Numerator'
    },
    'learningRateDenominator': {
        'description': 'The number of data samples used in conjunction with Learning Rate Numerator.',
        'type': 'integer',
        'default': '10000',
        'minimum': '1',
        'order': '6',
        'group': 'Model Tuning',
        'displayName': 'Learning Rate Denonminator'
    },
    'autoStreamingWindow': {
        'description': 'The number of data samples used in conjunction with Learning Rate Numerator.',
        'type': 'enumeration',
        'options': [ 'Automatic', 'Manually Configured' ],
        'default': 'Automatic',
        'order': '7',
        'group': 'Model Tuning',
        'displayName': 'Streaming Window Configuration'
    },
    'streamingWindowSize': {
        'description': 'The number of data samples to consider when... .',
        'type': 'integer',
        'default': '1',
        'minimum': '1',
        'order': '8',
        'group': 'Model Tuning',
        'validity': 'autoStreamingWindow == "Manually Configured"',
        'displayName': 'Streaming Window Size'
    },
    'username': {
        'description': 'Amber API username.',
        'type': 'string',
        'default': '',
        'order': '9',
        'group': 'Licence',
        'displayName': 'Username'
    },
    'password': {
        'description': 'Amber API password.',
        'type': 'password',
        'default': '',
        'order': '10',
        'group': 'Licence',
        'displayName': 'Password'
    },
    'server': {
        'description': 'Amber API server URL.',
        'type': 'string',
        'default': '',
        'order': '11',
        'group': 'Licence',
        'displayName': 'Server'
    },
    'enable': {
        'description': 'Enable/Disable filter plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'group': 'Naming',
        'order': '12'
    }
}

def plugin_info():
    """ Returns information about the plugin

    Args:

    Returns:
        dict: plugin information
    """
    return {
        'name': PLUGIN_NAME,
        'version': '3.0.0-1-g8504afd',
        'mode': 'none',
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG,
    }

AMBER_LICENSE_FILE = os.path.expanduser('~/.Amber.license')

def _create_or_update_license(username, password, server):
    """
    Creates or updates a license file with the specified user credentials and server information.

    Args:
        username (str): The username to be included in the license file.
        password (str): The password associated with the username.
        server (str): The server address for which the license is valid.
    Returns:
        none
    This function constructs a license configuration dictionary containing the provided
    username, password, and server information. It then writes this configuration to
    the specified license file (AMBER_LICENSE_FILE) in JSON format. If the file already exists,
    it will be overwritten with the new values.
    """
    license_config = {
        "default": {
            "username": username,
            "password": password,
            "server": server
        }
    }
    with open(AMBER_LICENSE_FILE, 'w+') as f:
        json.dump(license_config, f)

def plugin_init(config, ingest_ref, callback):
    """Initialise the plugin.

    Args:
        config: JSON configuration document for the Filter plugin configuration category.
        ingest_ref: Filter ingest reference.
        callback: Filter callback.

    Returns:
        data: JSON object to be used in future calls to the plugin.
    """
    # Create or update the license using provided credentials
    # Liecnse file is created in the user's home directory
    # License file is used to store the user credentials and server information
    # License file is created before the client is created
    _create_or_update_license(
                config['username']['value'],
                config['password']['value'],
                config['server']['value']
            )
    handle = _prepare_plugin_handle(config, ingest_ref, callback)
    try:
        if 'client' in handle and handle['client'] is not None:
            _configure_client(handle)
    except:
        _LOGGER.error(f"Failed to configure Amber client.")
        handle['client'] = None
    return handle

def _reconfigure_license(handle, new_config):
    """
    Reconfigures the license file with the specified user credentials and server information.

    Args:
        handle (dict): The handle returned by the plugin initialisation call.
        new_config (dict): The new configuration values to be used in the license file.
    Returns: none  
    """
    # Check if license have changed
    license_changed = (
        new_config['username']['value'] != handle['username']['value'] or
        new_config['password']['value'] != handle['password']['value'] or
        new_config['server']['value'] != handle['server']['value']
    )

    # Configure client based on whether license have changed
    if license_changed:
        # Create or update the license using provided credentials
        _LOGGER
        _create_or_update_license(
            new_config['username']['value'],
            new_config['password']['value'],
            new_config['server']['value']
        ) 
        # Recreate the client, as the license has changed
        handle['client'] = None
        _LOGGER.info("License has changed. Recreating client.")
    else:
        # License has not changed, so no need to recreate the client
        _LOGGER.info("License has not changed. Skipping license reconfiguration.")  

def plugin_reconfigure(handle, new_config):
    """Reconfigures the plugin.

    Args:
        handle: Handle returned by the plugin initialisation call.
        new_config: JSON object representing the new configuration category for the category.

    Returns:
        new_handle: New handle to be used in the future calls.
    """
    _LOGGER.info("Old config {} \n new config {} for {} plugin".format(handle, new_config, PLUGIN_NAME))

    _reconfigure_license(handle, new_config)

    new_handle = _prepare_plugin_handle(
        new_config,
        handle['ingestRef'],
        handle['callback'],
        client= handle['client'] if 'client' in handle else None
    )
    if 'client' in new_handle and new_handle['client'] is not None:
        try:
            _configure_client(new_handle)
            new_handle['client'].reconfigureSensors()
        except Exception as ex:
            _LOGGER.error(f"Failed to configure Amber client." + str(ex))  
            new_handle['client'] = None
    else:
        _LOGGER.warning("No client found in new_handle. Skipping reconfiguration.")
    return new_handle


def _prepare_plugin_handle(config, ingest_ref, callback, client=None):
    """Prepares plugin handle with the given configuration.

    Args:
        config: Configuration data.
        ingest_ref: Ingest reference.
        callback: Callback function.
        client: Optional client instance to reuse.

    Returns:
        data: Prepared data dictionary.
    """
    handle = copy.deepcopy(config)
    handle['callback'] = callback
    handle['ingestRef'] = ingest_ref
    try:
        handle['client'] = client or AmberClient()
    except Exception as ex:
        _LOGGER.error(f"Failed to create Amber client." + str(ex))
        handle['client'] = None
    return handle

def _configure_client(handle):
    """Configures the client with the given handle.

    Args:
        data: Plugin handle containing the client and configuration values.
    """
    client = handle['client']

    # Set client configuration parameters
    client.getSensors()
    client.setLearningSamples(handle['learningSamples']['value'])
    client.setModelName(handle['modelName']['value'])
    client.setLearningRateNumerator(handle['learningRateNumerator']['value'])
    client.setLearningRateDenominator(handle['learningRateDenominator']['value'])
    client.setAutoTuningSamples(handle['autoTuningSamples']['value'])

    # Configure streaming window size based on settings
    if handle['autoStreamingWindow']['value'] == 'Manually Configured':
        client.setStreamingWindowSize(handle['streamingWindowSize']['value'])
    else:
        client.setStreamingWindowSize('-1')

def plugin_shutdown(handle):
    """ Shutdown the plugin doing required cleanup.

    Args:
        handle: handle returned by the plugin initialisation call
    """
    handle['callback'] = None
    handle['ingestRef'] = None
    _LOGGER.info('{} filter plugin shutdown.'.format(PLUGIN_NAME))

def plugin_ingest(handle, data):
    """ Modify readings data and pass it onward

    Args:
        handle: handle returned by the plugin initialisation call
        data:   readings data
    """
    try:
        if handle['enable']['value'] == 'false':
            # Filter not enabled, just pass data onwards
            return

        # Filter is enabled: compute for each reading
        if 'client' in handle and handle['client'] is not None:
            client = handle['client']
        else:
            try:
                client = AmberClient()
                handle['client'] = client
                _configure_client(handle)
                version = client.get_version()
                _LOGGER.info("Connected to Boon Amber version {}.".format(version))
            except Exception as ex:
                _LOGGER.error(f"Failed to create Amber client." + str(ex))
                handle['client'] = None
                return
        client.ingest(data)
    finally:
        filter_ingest.filter_ingest_callback(handle['callback'],  handle['ingestRef'], data)
