import logging
import numbers
import decimal
import re
from foglamp.common import logger
from boonamber import AmberV1Client, AmberCloudError, AmberUserError, ApiException

_LOGGER = logger.setup(__name__, level=logging.WARN)

__author__ = "Ashish Jabble, Mark Riddoch"
__copyright__ = "Copyright (c) 2024 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


class AmberClient(object):
    """ An amber client to the Boon Logic """
    __slots__ = ['client', 'sensors', 'features', 'learningSamples', 'modelName',
            'autoTuningSamples', 'learningRateNumerator', 'learningRateDenominator',
            'streamingWindowSize', 'datapointMap', 'non_alphanumeric_re']

    def __init__(self):
        # Use ~/.Amber.license file and "default" profile
        # We may need to save file under FOGLAMP_DATA so that we have control of custom fields in Plugin configuration
        self.client = AmberV1Client()
        self.sensors = {}
        self.features = {}
        self.datapointMap = {}
        self.learningSamples = 100000
        self.modelName = 'amber'
        self.autoTuningSamples = 10000
        self.learningRateNumerator = 10
        self.learningRateDenominator = 10000
        self.streamingWindowSize = 1
        self.non_alphanumeric_re = re.compile('[^0-9A-Za-z]')

    def get_version(self) -> dict:
        """ version information for the API """
        try:
            version_info = self.client.get_version()
            return version_info
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to get version.")

    def get_models(self) -> dict:
        """ List all models belonging to current user """
        try:
            models = self.client.get_models()
            return models.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to get all models.")

    def get_model_info(self, model_id: str) -> dict:
        """ Return metadata for the specified model

        Args:
            model_id: Model ID (required)

        Returns:
            Dict
        """
        try:
            model = self.client.get_model(model_id)
            return model.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to get metadata for the specified model with ID: {}.".format(model_id))

    def update_label(self, model_id: str, label: str) -> dict:
        """ Update metadata for the specified model

        Args:
            model_id: Model ID (required)
            label: updates to apply (required)

        Returns:
            Dict
        """
        try:
            model = self.client.update_label(model_id, label)
            return model.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to update metadata for the specified model with ID: {}.".format(model_id))

    def post_config(self, model_id: str, features: list) -> dict:
        """ Configure the specific model. Wipes all progress and puts the model in the Buffering state

        Args:
            model_id: Model ID (required)
            features: Configuration to apply

        Returns:
            Dict
        """
        try:
            param = boonamber.PostConfigRequest(features=features)
            config = self.client.post_config(model_id, param)
            return config.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to configure the specified model with ID: {}.".format(model_id))

    def get_config(self, model_id: str) -> dict:
        """ Get the configuration of the specified model

        Args:
            model_id:   Model ID (required)

        Returns:
            Dict
        """
        try:
            config = self.client.get_config(model_id)
            return config.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to get configuration of the specified model with ID: {}.".format(model_id))

    def post_data(self, model_id: str, data: str, save_image: bool) -> dict:
        """ Send data to the specified model, and get back the resulting analytics and model status

        Args:
            model_id:   Model ID (required)
            data:       Data vector or vectors as flattened list of comma separated values (required)
            save_image: Determine whether to save the model (only applies to on prem)
        Returns:
            Dict
        """
        try:
            results = self.client.post_data(model_id, data=data, save_image=save_image)
            return results.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to send data to the specified model with ID: {}.".format(model_id))

    def get_status(self, model_id: str) -> dict:
        """ Get the current state and learning progress of the specified model

        Args:
            model_id: Model ID (required)

        Returns:
            Dict
        """
        try:
            status = self.client.get_status(model_id)
            return status.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Unable to get the current state of the specified model with ID: {}.".format(model_id))

    def get_nano_status(self, model_id: str) -> dict:
        """ Get the current nano status of the specified model

        Args:
            model_id: Model ID (required)

        Returns:
            Dict
        """
        try:
            status = self.client.get_nano_status(model_id)
            return status.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Unable to get the nano status of the specified model with ID: {}.".format(model_id))

    def delete_model(self, model_id: str) -> dict:
        """ Permanently delete the specified model

        Args:
            model_id: Model ID (required)

        Returns:
            Dict
        """
        try:
            response = self.client.delete_model(model_id)
            return response.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Unable to delete the specified model with ID: {}.".format(model_id))

    def copy_model(self, model_id: str, label: str) -> dict:
        """ Copy a model and return the new model information

        Args:
            model_id: Model ID (required)
            label: label for new model

        Returns:
            Dict
        """
        try:
            copy_parameters = {
                'new_model_name': label,
                # Add other parameters as needed, e.g., description
            }
            response = self.client.copy_model(model_id, **copy_parameters)
            return response.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Unable to copy the model with ID: {}.".format(model_id))

    def get_pretrain(self, model_id: str) -> dict:
        """ Get the pretraining status of the specified model

        Args:
            model_id: Model ID (required)

        Returns:
            Dict
        """
        try:
            status = self.client.get_pretrain(model_id)
            return status.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Unable to get pretraining status of the specified model with ID: {}.".format(model_id))

    def post_pretrain(self, model_id: str, data: list, chunk_size: int, block: bool) -> dict:
        """ pretrain model with existing dataset

        Args:
            model_id: Model ID (required)
            data: data to process
            chunk_size: number of portions to send the data over
            block: wait until pretraining finishes before returning

        Returns:
            Dict
        """
        try:
            response = self.client.post_pretrain(model_id=model_id, data=data, chunk_size=chunk_size, block=block)
            return response.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to pretrain model with ID: {}.".format(model_id))

    def put_data(self, model_id: str, vector: list) -> dict:
        """ update fusion vector and get back results

        Args:
            model_id: Model ID (required)
            vector: fusion vector list (required)

        Returns:
            Dict
        """
        try:
            body = boonamber.PutDataRequest(vector=vector)
            response = self.client.put_data(model_id, body=body)
            return response.to_dict()
        except ApiException as ex:
            msg = str(ex)
            _LOGGER.error(msg, "Failed to update fusion vector for the model with ID: {}.".format(model_id))

    def setModelName(self, modelName):
        """ Set the name of the model

        Args:
            modelName: The name of the model

        Rebuild the list of sensors and features previously discovered
        """
        self.modelName = modelName
        if len(self.sensors) != 0:
            self.sensors.clear()
        if len(self.features) != 0:
            self.features.clear()
        self.getSensors()

    def setLearningSamples(self, samples):
        """ Set the number of samples needed for learning

        Args:
            samples: The number of samples required
        """
        self.learningSamples = int(samples)

    def setAutoTuningSamples(self, samples):
        """ Set the number of samples needed for auto tuning

        Args:
            samples: The number of samples required
        """
        self.autoTuningSamples = int(samples)

    def setLearningRateNumerator(self, samples):
        """ Set the number of samples needed for auto tuning

        Args:
            samples: The number of samples required
        """
        self.learningRateNumerator = int(samples)

    def setLearningRateDenominator(self, samples):
        """ Set the number of samples needed for auto tuning

        Args:
            samples: The number of samples required
        """
        self.learningRateDenominator = int(samples)

    def setStreamingWindowSize(self, samples):
        """ Set the number of samples needed for auto tuning

        Args:
            samples: The number of samples required
        """
        self.streamingWindowSize = int(samples)

    def deleteSensors(self):
        """ Remove all sensors for this model
        """
        sens = self.client.list_sensors()
        for id in sens:
            model = self.modelName + ':'
            fullname = sens[id]
            if model in fullname:
                self.client.delete_sensor(id)

    def getSensors(self):
        """ Fetch the sensor from the Amber instance
        """
        sens = self.client.list_sensors()
        _LOGGER.info("Sensor list is {}.".format(sens))
        modelName = self.modelName + ':'
        for id in sens:
            fullname = sens[id]
            if fullname.find(modelName) != -1:
                _LOGGER.info("Model/Asset {}/{} id {}".format(modelName, fullname, id))
                try:
                    config = self.client.get_config(id)
                    features = config['features']
                    labels = []
                    for f in features:
                        label = f['label']
                        labels.append(label)
                    self.features[fullname] = labels
                    _LOGGER.debug("asset: {} has features {}".format(fullname, labels))
                    self.sensors[fullname] = id
                except AmberCloudError as e:
                    _LOGGER.error("Failed to get sensor {}, {}.".format(fullname, e))
                    self.client.delete_sensor(id)
                    _LOGGER.warn("Deleted sensor {}".format(fullname))

    def reconfigureSensors(self):
        """Dynamically reconfigure the model.

        Fetch the model configuration and check if any of the configuration
        options are different from the current settings. If needed, reconfigure the sensors.
        """
        model_exists = False
        sensors = self.client.list_sensors()

        for sensor_id, fullname in sensors.items():
            if fullname.startswith(f"{self.modelName}:"):
                model_exists = True
                _LOGGER.info("Reconfigure sensor %s", fullname)

                # Fetch current configuration
                current_config = self.client.get_config(sensor_id)
                _LOGGER.info(current_config)
                if self._needs_reconfiguration(current_config):
                    self._reconfigure_sensor(sensor_id, fullname, current_config)

        if not model_exists:
            # Sensor does not exist, so log and handle appropriately
            _LOGGER.fatal("As a result of reconfiguration, a new model needs to be created.")


    def ingest(self, data):
        """Ingest data from upstream.

        Create a sensor for each asset in the readings. Every numeric datapoint
        within the asset becomes a feature for that sensor. Sensors are created only once.
        The state and anomaly index (SI) are added to each asset as a new data point
        and sent onwards by the filter.

        Args:
            data: A list of FogLAMP readings.

        Returns:
            None
        """
        # Create sensors and/or datapoint map if needed
        for reading in data:
            asset = reading['asset']
            fullname = f"{self.modelName}:{asset}"
            if fullname not in self.sensors:
                self._create_sensor(fullname, reading['readings'])
            elif fullname not in self.datapointMap:
                self.datapointMap[fullname] = {}
                readings = reading['readings']
                for key, value in readings.items():
                    sensor = self._map_sensor_name(key)
                    if isinstance(value, numbers.Number):
                       self.datapointMap[fullname][sensor] = key

        # Prepare and stream data to sensors
        sensor_data = self._prepare_sensor_data(data)
        self._stream_sensor_data(sensor_data)


    # ----------------------------------------
    # Helper Methods
    # ----------------------------------------

    def _needs_reconfiguration(self, current_config):
        """Determine if a sensor needs reconfiguration based on current settings."""
        streaming_window_size = self._get_streaming_window_size()
        return (
            current_config['learningMaxSamples'] != self.learningSamples
            or current_config['learningRateNumerator'] != self.learningRateNumerator
            or current_config['learningRateDenominator'] != self.learningRateDenominator
            or current_config['streamingWindowSize'] != streaming_window_size
        )


    def _reconfigure_sensor(self, sensor_id, fullname, current_config):
        """Reconfigure the specified sensor if needed."""
        try:
            features = current_config['features']
            streaming_window_size = self._get_streaming_window_size()

            _LOGGER.info(
                "Reconfiguring sensor %s with samples=%s, max_samples=%s, numerator=%s, denominator=%s, window_size=%s",
                fullname, self.autoTuningSamples, self.learningSamples, self.learningRateNumerator,
                self.learningRateDenominator, streaming_window_size
            )

            new_config = self.client.configure_sensor(
                sensor_id,
                feature_count=len(features),
                features=features,
                samples_to_buffer=self.autoTuningSamples,
                learning_max_samples=self.learningSamples,
                learning_rate_numerator=self.learningRateNumerator,
                learning_rate_denominator=self.learningRateDenominator,
                streaming_window_size=streaming_window_size
            )
            _LOGGER.debug("Reconfigured sensor %s. New configuration: %s", fullname, new_config)
        except ApiException as ex:
            _LOGGER.error("Failed to reconfigure sensor %s. Error: %s", fullname, str(ex))


    def _create_sensor(self, fullname, readings):
        """Create a new sensor for the given asset."""
        try:
            _LOGGER.info("Creating new sensor for %s.", fullname)
            sensor_id = self.client.create_sensor(fullname)
            self.sensors[fullname] = sensor_id

            self.datapointMap[fullname] = {}
            features, keys = self._extract_features(readings, self.datapointMap[fullname])
            self.features[fullname] = keys

            streaming_window_size = self._get_streaming_window_size(keys)

            _LOGGER.info(
                "Configuring sensor %s with samples=%s, max_samples=%s, numerator=%s, denominator=%s, window_size=%s",
                fullname, self.autoTuningSamples, self.learningSamples, self.learningRateNumerator,
                self.learningRateDenominator, streaming_window_size
            )

            config = self.client.configure_sensor(
                sensor_id,
                feature_count=len(features),
                features=features,
                samples_to_buffer=self.autoTuningSamples,
                learning_max_samples=self.learningSamples,
                learning_rate_numerator=self.learningRateNumerator,
                learning_rate_denominator=self.learningRateDenominator,
                streaming_window_size=streaming_window_size
            )
            _LOGGER.debug("Sensor %s configured with: %s", fullname, config)
        except (ApiException, AmberCloudError, AmberUserError) as ex:
            _LOGGER.error("Failed to create or configure sensor %s. Error: %s", fullname, str(ex))


    def _extract_features(self, readings, datapointMap):
        """Extract features from the readings."""
        features = []
        sensors = []
        for key, value in readings.items():
            sensor = self._map_sensor_name(key)
            if isinstance(value, numbers.Number):
                _LOGGER.info("Adding feature %s to asset.", sensor)
                features.append({"label": sensor, "submitRule": "submit"})
                sensors.append(sensor)
                datapointMap[sensor] = key
        return features, sensors


    def _get_streaming_window_size(self, keys=None):
        """Determine the streaming window size based on settings."""
        if self.streamingWindowSize == -1:
            return 25 if keys and len(keys) == 1 else 1
        return self.streamingWindowSize


    def _prepare_sensor_data(self, data):
        """Prepare data to be streamed to sensors."""
        sensor_data = {}
        for reading in data:
            asset = reading['asset']
            fullname = f"{self.modelName}:{asset}"
            if fullname in self.sensors:
                if fullname not in sensor_data:
                    sensor_data[fullname] = []
                sd = sensor_data[fullname]

                if fullname not in self.features:
                    _LOGGER.error("No feature set defined for %s.", fullname)
                    continue

                # Collect feature values
                values = [reading['readings'][self.datapointMap[fullname][key]] for key in self.features[fullname]]
                sd.append({'reading': reading['readings'], 'data': values})
            else:
                _LOGGER.error("No sensor or asset %s found. Full sensor name: %s.", asset, fullname)
        return sensor_data


    def _stream_sensor_data(self, sensor_data):
        """Stream data to sensors and handle the response."""
        for fullname, data in sensor_data.items():
            all_data = [item['data'] for item in data]
            try:
                response = self.client.stream_sensor(self.sensors[fullname], all_data)
                _LOGGER.debug("Stream response: %s", response)

                # Update readings with response details
                for offset, item in enumerate(data):
                    reading = item['reading']
                    reading['amber:state'] = response['state']
                    reading['amber:progress'] = response['progress']
                    reading['amber:clusterCount'] = response['clusterCount']
                    reading['amber:ID'] = response['ID'][offset]
                    reading['amber:SI'] = response['SI'][offset]
                    reading['amber:AD'] = response['AD'][offset]
                    reading['amber:AH'] = response['AH'][offset]
                    reading['amber:AM'] = response['AM'][offset]
                    reading['amber:AW'] = response['AW'][offset]
                    reading['amber:NI'] = response['NI'][offset]
                    reading['amber:NS'] = response['NS'][offset]
                    reading['amber:NW'] = response['NW'][offset]
                    reading['amber:OM'] = response['OM'][offset]
                    reading['amber:CS'] = 100.0 - (50.0 * response['NW'][offset])
                    if response['message'] != '':
                        reading['amber:message'] = response['message']

                    # Handle root cause data if necessary
                    ID = response['ID'][offset]
                    if ID < 0:
                        self._handle_root_cause(reading, fullname, ID)
                    else:
                        self._add_empty_root_cause(reading, fullname)
            except ApiException as ex:
                _LOGGER.error("Failed to stream data to sensor %s. Error: %s", fullname, str(ex))


    def _handle_root_cause(self, reading, fullname, ID):
        """Handle root cause analysis for a given sensor."""
        try:
            root_cause = self.client.get_root_cause(self.sensors[fullname], id_list=[ID])
            sensor = 0
            for key in self.features[fullname]:
                value = root_cause[0][sensor]
                sensorName = self.datapointMap[fullname][key]
                reading['amber:RCA_' + sensorName] = value
                sensor += 1
        except ApiException as ex:
            _LOGGER.error("Failed to retrieve root cause for sensor %s. Error: %s", fullname, str(ex))

    def _add_empty_root_cause(self, reading, fullname):
        """No anomaly occured so set an empty root cause"""
        for key in self.features[fullname]:
            sensor = self.datapointMap[fullname][key]
            reading['amber:RCA_' + sensor] = 0.0

    def _map_sensor_name(self, name):
        """Map any non-alphanumeric characters to an underscore"""
        return self.non_alphanumeric_re.sub('_', name)
